/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : Wed Mar 27 21:51:30 EET 2002
    copyright            : (C) 2002 by Alin Nastac
    email                : mrness@gentoo.ro
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

struct tagRouteEntry
{
	unsigned long addr;
	unsigned long mask;
	unsigned bits;
	struct tagRouteEntry* pNext;
};

char *g_pszSourceFirstChain = NULL, *g_pszDestinationFirstChain = NULL;
unsigned long	g_nMark = 0;
char sLine[1024];
int g_bFilterPrivateIPs = 1;
int g_bVerbose = 0;

int checkbyte(unsigned long x)
{
	return x >=0 && x <= 0xff;
}

int addr_intersected(struct tagRouteEntry *pEntry1, struct tagRouteEntry *pEntry2, unsigned bits)
{
		return pEntry1->bits >= bits && pEntry2->bits >= bits &&
				(pEntry1->addr ^ pEntry2->addr) >> (32 - bits) == 0;
}

void merge_included_prefixes(struct tagRouteEntry *pCurrent, unsigned long addr, unsigned long mask, unsigned b)
{
	pCurrent->addr = addr;
	pCurrent->mask = mask;
	pCurrent->bits = b;
	while(pCurrent->pNext &&
		pCurrent->pNext->bits >= b && (pCurrent->pNext->addr & mask) == addr)
	{
		struct tagRouteEntry *pToRemove = pCurrent->pNext;
		pCurrent->pNext = pCurrent->pNext->pNext;
		free(pToRemove);
	}
}

void printhelp(char* pszName)
{
		printf("usage: %s [ -v ] [ -p ] [ -s initial_src_chain ] [ -d initial_dst_chain ] [ -m number ]\n"
			"\t-v\tverbose (show number of reduced prefixes when you don't use -s or -d)"
			"\t-p\tdon't filter private IP prefixes\n"
			"\t-s\tiptables chains by source IP\n"
			"\t-d\tiptables chains by destination IP\n"
			"\t-m\tmark (mandatory if you specify -s or -d)\n\n"
			"Without any parameters it will just print the reduced prefixes\n",
			pszName);
}

int add_mark_rule(char* pszDir, char* pszChain, struct tagRouteEntry *pEntry)
{
		printf("-A %s %s %lu.%lu.%lu.%lu/%u -j MARK --set-mark %ld\n",
				pszChain, pszDir,
				pEntry->addr>>24, (pEntry->addr>>16)&0xff,
				(pEntry->addr>>8)&0xff, pEntry->addr&0xff, pEntry->bits,
				g_nMark);
		return EXIT_SUCCESS;
}

int add_subchain_rule(char* pszDir, char* pszChain, char* pszSubchain, unsigned long addr, unsigned bits)
{
		printf(":%s - [0:0]\n", pszSubchain);
		printf("-A %s %s %lu.%lu.%lu.%lu/%u -j %s\n",
				pszChain, pszDir,
				addr>>24, (addr>>16)&0xff,
				(addr>>8)&0xff, addr&0xff, bits,
				pszSubchain);
		return EXIT_SUCCESS;
}

int buildchain(char* pszDir, char* pszChain, unsigned bits, struct tagRouteEntry *pFirst, struct tagRouteEntry *pLast)
{
	struct tagRouteEntry *pCurrent = NULL;
	int count, ret = 0;

	while(pFirst)
	{
		count = 1;
		pCurrent = pFirst;

		while(pCurrent != pLast && pCurrent->pNext)
		{
			if(addr_intersected(pFirst, pCurrent->pNext, bits))
			{
				count++;
				pCurrent = pCurrent->pNext;
			}
			else
				break;
		}

		if(count <= 1)
		{
			int i;
			for(i=0; i<count; i++)
			{
				ret = add_mark_rule(pszDir, pszChain, pFirst);
				if(ret)
					return ret;
				pFirst = pFirst->pNext;
			}
		}
		else
		{
			char pszSubchain[255];
			unsigned long addr = (pFirst->addr & (0xFFFFFFFF << (32-bits)));
			sprintf(pszSubchain, "%s_%d", pszChain, (int)(addr >> (32-bits)) & 0xFF);
						
			ret = add_subchain_rule(pszDir, pszChain, pszSubchain, addr, bits);
			if(ret)
				return ret;

			ret = buildchain(pszDir, pszSubchain, bits + 8, pFirst, pCurrent);
			if(ret)
				return ret;
		}

		if(pCurrent == pLast)
			break;
		else
			pFirst = pCurrent->pNext;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	unsigned long a1, a2, a3, a4, b;
	unsigned long addr, mask;
	struct tagRouteEntry *pRouteTable = NULL, *pCurrent = NULL;
	int x, ret = EXIT_SUCCESS;
	for(x=1; x<argc; x++)
	{
		if(strcmp(argv[x], "-v") == 0 )
			g_bVerbose = 1;
		else if(strcmp(argv[x], "-p") == 0 )
			g_bFilterPrivateIPs = 0;
		else if(strcmp(argv[x], "-s") == 0 && x < argc - 1)
			g_pszSourceFirstChain = argv[++x];
		else if(strcmp(argv[x], "-d") == 0	&& x < argc - 1)
			g_pszDestinationFirstChain = argv[++x];
		else if(strcmp(argv[x], "-m") == 0 && x < argc - 1)
			g_nMark = atoi(argv[++x]);
		else
		{
			printhelp(argv[0]);
			return strcmp(argv[x], "--help") ? 1 : 0;
		}
	}
	if(g_pszSourceFirstChain || g_pszDestinationFirstChain)
	{
		if(!g_nMark)
		{
			printhelp(argv[0]);
			return 1;
		}
		else
		{
		 	/* Begin generation of iptables-restore input 
			by flushing and removing mangle chains which begins with the special prefix */
			time_t now = time(NULL);
			FILE* fd;
			printf("# Generated by mipclasses on %s*mangle\n", ctime(&now));

			if(g_pszSourceFirstChain && g_pszDestinationFirstChain)
				sprintf(sLine, "iptables-save -t mangle | egrep '^:(%s|%s)_'",
					g_pszSourceFirstChain, g_pszDestinationFirstChain);
			else
				sprintf(sLine, "iptables-save -t mangle | egrep '^:%s_'",
					(g_pszSourceFirstChain ? g_pszSourceFirstChain : g_pszDestinationFirstChain));
			if(g_pszSourceFirstChain)
				printf("-F %s\n", g_pszSourceFirstChain);
			if(g_pszDestinationFirstChain)
				printf("-F %s\n", g_pszDestinationFirstChain);

			fd = popen(sLine, "r");
			if(fd == NULL)
				perror("cannot run iptables-save");
			while(fgets(sLine,1023,fd))
			{
				char chain[256];
				if(sscanf(sLine+1, "%s", chain) == 1)
					printf("-F %s\n-X %s\n", chain, chain);
			}
			pclose(fd);
		}
	}
	while(1)
	{
		if(fgets(sLine,1023,stdin) == NULL)
			break;
		x = 0;
		while(sLine[x] && (sLine[x] < '0' || sLine[x] > '9')) x++;
		x = sscanf(sLine+x, "%lu.%lu.%lu.%lu/%lu", &a1, &a2, &a3, &a4, &b);
		switch(x)
		{
			case 4:
				if(a1 >= 192) b = 24;
				else if (a1 >= 128) b = 16;
				else b = 8;
				
				if( a4 != 0 || ( b <= 16 && a3 != 0) || ( b <= 8 && a2 != 0 ))
					break; /* this IP does not represent a network address */
			case 5:
				if(b > 0 && b <= 32 && a1 != 0 && /*no sense in running this program if you want to add 0.0.0.0/0; 0. is also filtered */
					(
						!g_bFilterPrivateIPs ||
						(
							a1 != 10 &&
							( a1 != 172 || a2 < 16 || a2 > 31 ) &&
							( a1 != 192 || a2 != 168 )
						)
					) &&
					checkbyte(a1) && checkbyte(a2) && checkbyte(a3) && checkbyte(a4))
				{
					addr = (a1 << 24) | (a2 << 16) | (a3 << 8) | a4;
					mask = 0xFFFFFFFF << (32-b);
					addr &= mask;
					
					/* Merge this prefix with whatever imbrincated prefixes do we have until now */
					if (pRouteTable == NULL)
					{
						/* If it is the first prefix then initialize list with this */
						pRouteTable = malloc(sizeof(struct tagRouteEntry));
						if(pRouteTable == NULL)
							return 1;
						pCurrent = pRouteTable;
						pCurrent->addr = addr;
						pCurrent->mask = mask;
						pCurrent->bits = b;
						pCurrent->pNext = NULL;
					}
					else
					{
						/* Find the last prefix with net address <= with this one */
						pCurrent = pRouteTable;
						while(pCurrent->pNext && pCurrent->pNext->addr <= addr)
							pCurrent = pCurrent->pNext;
						
						if(b >= pCurrent->bits && (addr & pCurrent->mask) == pCurrent->addr)
							break; /* it is included in pCurrent prefix so it could be ignored */
						else if(pCurrent->bits >= b && (pCurrent->addr & mask) == addr)
						{
							/* pCurrent prefix is included in this one; set pCurrent to this one
							   merge any following prefixes that match this prefix; it is already sorted 
								YES, I know what you're thinking!!! You think that case will never occur but
								you forget about (pCurrent == pRouteTable) case*/
							merge_included_prefixes(pCurrent, addr, mask, b);
						}
						else if(pCurrent->pNext &&
							pCurrent->pNext->bits >= b && (pCurrent->pNext->addr & mask) == addr)
						{
							/* pCurrent->pNext prefix is included in this one; set pCurrent to pCurrent->pNext
							   Note: this prefix cannot be included in pCurent->pNext since pCurrent->pNext->addr > addr*/
							pCurrent = pCurrent->pNext;
							/* pCurrent prefix is included in this one; set pCurrent to this one
							   merge any following prefixes that match this prefix; it is already sorted	*/
							merge_included_prefixes(pCurrent, addr, mask, b);
						}
						else
						{
							struct tagRouteEntry *pNew = NULL;
							pNew = malloc(sizeof(struct tagRouteEntry));
							if(pNew == NULL)
								return 2;
							pNew->addr = addr;
							pNew->mask = mask;
							pNew->bits = b;
							if(pCurrent == pRouteTable &&	pCurrent->addr > addr)
							{
								pNew->pNext = pRouteTable;
								pRouteTable = pNew;
							}
							else
							{
								pNew->pNext = pCurrent->pNext;
								pCurrent->pNext = pNew;
							}
						}
					}
				}
		}
	}

	/* Compose neighbour IP prefixes; exit from main loop when no composition has occured */
	do
	{
		x = 0;
		pCurrent =	pRouteTable;
		while(pCurrent && pCurrent->pNext)
		{
			if(pCurrent->bits == pCurrent->pNext->bits &&
				(1 << (32 - pCurrent->bits)) == (pCurrent->addr ^ pCurrent->pNext->addr))
			{
				struct tagRouteEntry *pTemp = pCurrent->pNext->pNext;
				free(pCurrent->pNext);
				pCurrent->pNext = pTemp;
				pCurrent->bits--;
				pCurrent->mask <<= 1;
				pCurrent->addr &= pCurrent->mask;
				x = 1;
			}
			else
				pCurrent = pCurrent->pNext;
		}
	} while(x);

	if(g_pszSourceFirstChain || g_pszDestinationFirstChain)
	{
		/* Continue creation of the iptables-restore input. 
		This time create the new trees of chains */
		if(g_pszSourceFirstChain)
			ret = buildchain("-s", g_pszSourceFirstChain, 8, pRouteTable, NULL);
		if(ret == EXIT_SUCCESS &&	g_pszDestinationFirstChain)
			ret = buildchain("-d", g_pszDestinationFirstChain, 8, pRouteTable, NULL);
		if(ret == EXIT_SUCCESS)
		{
			time_t now = time(NULL);
			printf("COMMIT\n# Completed on %s", ctime(&now));
		}
	}
	else
	{
		/* Display the reduced IP prefixes */
		int count = 0;
		pCurrent =	pRouteTable;
		while(pCurrent)
		{
			count++;
			printf("%lu.%lu.%lu.%lu/%u\n",
					pCurrent->addr>>24, (pCurrent->addr>>16)&0xff,
					(pCurrent->addr>>8)&0xff, pCurrent->addr&0xff, pCurrent->bits);
			pCurrent = pCurrent->pNext;
		}
		if(g_bVerbose)
			printf("------------------------------\n%d reduced prefixes\n", count);
	}

	/* Free memory */
	pCurrent =	pRouteTable;
	while(pCurrent)
	{
		pCurrent = pCurrent->pNext;
		free(pRouteTable);
		pRouteTable = pCurrent;
	}
	return ret;
}
